@@ -6,14 +6,17 @@ use warnings;
 use base qw/DBIx::Class::Storage::DBI/;
 use mro 'c3';
 
-use POSIX 'strftime';
-use File::Copy;
-use File::Spec;
+__PACKAGE__->sql_maker_class('DBIx::Class::SQLMaker::SQLite');
+__PACKAGE__->sql_limit_dialect ('LimitOffset');
+__PACKAGE__->sql_quote_char ('"');
+__PACKAGE__->datetime_parser_type ('DateTime::Format::SQLite');
 
-__PACKAGE__->sql_maker_class('DBIx::Class::SQLAHacks::SQLite');
+sub backup {
+
+  require File::Spec;
+  require File::Copy;
+  require POSIX;
 
-sub backup
-{
   my ($self, $dir) = @_;
   $dir ||= './';
 
@@ -25,39 +28,70 @@ sub backup
   {
     $dbname = $1 if($dsn =~ /^dbi:SQLite:(.+)$/i);
   }
-  $self->throw_exception("Cannot determine name of SQLite db file") 
+  $self->throw_exception("Cannot determine name of SQLite db file")
     if(!$dbname || !-f $dbname);
 
 #  print "Found database: $dbname\n";
 #  my $dbfile = file($dbname);
   my ($vol, $dbdir, $file) = File::Spec->splitpath($dbname);
 #  my $file = $dbfile->basename();
-  $file = strftime("%Y-%m-%d-%H_%M_%S", localtime()) . $file; 
+  $file = POSIX::strftime("%Y-%m-%d-%H_%M_%S", localtime()) . $file;
   $file = "B$file" while(-f $file);
 
   mkdir($dir) unless -f $dir;
   my $backupfile = File::Spec->catfile($dir, $file);
 
-  my $res = copy($dbname, $backupfile);
+  my $res = File::Copy::copy($dbname, $backupfile);
   $self->throw_exception("Backup failed! ($!)") if(!$res);
 
   return $backupfile;
 }
 
+sub _exec_svp_begin {
+  my ($self, $name) = @_;
+
+  $self->_dbh->do("SAVEPOINT $name");
+}
+
+sub _exec_svp_release {
+  my ($self, $name) = @_;
+
+  $self->_dbh->do("RELEASE SAVEPOINT $name");
+}
+
+sub _exec_svp_rollback {
+  my ($self, $name) = @_;
+
+  # For some reason this statement changes the value of $dbh->{AutoCommit}, so
+  # we localize it here to preserve the original value.
+  local $self->_dbh->{AutoCommit} = $self->_dbh->{AutoCommit};
+
+  $self->_dbh->do("ROLLBACK TRANSACTION TO SAVEPOINT $name");
+}
+
 sub deployment_statements {
-  my $self = shift;;
+  my $self = shift;
   my ($schema, $type, $version, $dir, $sqltargs, @rest) = @_;
 
   $sqltargs ||= {};
 
-  if (my $version = $self->_server_info->{dbms_version}) {
-    $sqltargs->{producer_args}{sqlite_version} = $version;
+  if (
+    ! exists $sqltargs->{producer_args}{sqlite_version}
+      and
+    my $dver = $self->_server_info->{normalized_dbms_version}
+  ) {
+    $sqltargs->{producer_args}{sqlite_version} = $dver;
   }
 
   $self->next::method($schema, $type, $version, $dir, $sqltargs, @rest);
 }
 
-sub datetime_parser_type { return "DateTime::Format::SQLite"; } 
+sub bind_attribute_by_data_type {
+  $_[1] =~ /^ (?: int(?:eger)? | (?:tiny|small|medium|big)int ) $/ix
+    ? do { require DBI; DBI::SQL_INTEGER() }
+    : undef
+  ;
+}
 
 =head2 connect_call_use_foreign_keys
 
