@@ -13,7 +13,6 @@ use warnings;
 use base 'DBIx::Class::Storage';
 use mro 'c3';
 
-use Carp::Clan qw/^DBIx::Class/;
 use List::Util 'first';
 use Scalar::Util 'blessed';
 use namespace::clean;
@@ -23,10 +22,11 @@ use namespace::clean;
 # {from} specs, aiding the RDBMS query optimizer
 #
 sub _prune_unused_joins {
-  my ($self) = shift;
-
+  my $self = shift;
   my ($from, $select, $where, $attrs) = @_;
 
+  return $from unless $self->_use_join_optimizer;
+
   if (ref $from ne 'ARRAY' || ref $from->[0] ne 'HASH' || ref $from->[1] ne 'ARRAY') {
     return $from;   # only standard {from} specs are supported
   }
@@ -39,7 +39,14 @@ sub _prune_unused_joins {
 
   my @newfrom = $from->[0]; # FROM head is always present
 
-  my %need_joins = (map { %{$_||{}} } (values %$aliastypes) );
+  my %need_joins;
+  for (values %$aliastypes) {
+    # add all requested aliases
+    $need_joins{$_} = 1 for keys %$_;
+
+    # add all their parents (as per joinpath which is an AoH { table => alias })
+    $need_joins{$_} = 1 for map { values %$_ } map { @$_ } values %$_;
+  }
   for my $j (@{$from}[1..$#$from]) {
     push @newfrom, $j if (
       (! $j->[0]{-alias}) # legacy crap
@@ -59,7 +66,7 @@ sub _adjust_select_args_for_complex_prefetch {
   my ($self, $from, $select, $where, $attrs) = @_;
 
   $self->throw_exception ('Nothing to prefetch... how did we get here?!')
-    if not @{$attrs->{_prefetch_select}};
+    if not @{$attrs->{_prefetch_selector_range}};
 
   $self->throw_exception ('Complex prefetches are not supported on resultsets with a custom from attribute')
     if (ref $from ne 'ARRAY' || ref $from->[0] ne 'HASH' || ref $from->[1] ne 'ARRAY');
@@ -70,7 +77,7 @@ sub _adjust_select_args_for_complex_prefetch {
   delete $outer_attrs->{$_} for qw/where bind rows offset group_by having/;
 
   my $inner_attrs = { %$attrs };
-  delete $inner_attrs->{$_} for qw/for collapse _prefetch_select _collapse_order_by select as/;
+  delete $inner_attrs->{$_} for qw/for collapse _prefetch_selector_range _collapse_order_by select as/;
 
 
   # bring over all non-collapse-induced order_by into the inner query (if any)
@@ -87,7 +94,9 @@ sub _adjust_select_args_for_complex_prefetch {
   # on the outside we substitute any function for its alias
   my $outer_select = [ @$select ];
   my $inner_select = [];
-  for my $i (0 .. ( @$outer_select - @{$outer_attrs->{_prefetch_select}} - 1) ) {
+
+  my ($p_start, $p_end) = @{$outer_attrs->{_prefetch_selector_range}};
+  for my $i (0 .. $p_start - 1, $p_end + 1 .. $#$outer_select) {
     my $sel = $outer_select->[$i];
 
     if (ref $sel eq 'HASH' ) {
@@ -100,31 +109,56 @@ sub _adjust_select_args_for_complex_prefetch {
     push @{$inner_attrs->{as}}, $attrs->{as}[$i];
   }
 
-  # construct the inner $from for the subquery
+  # construct the inner $from and lock it in a subquery
   # we need to prune first, because this will determine if we need a group_by below
   # the fake group_by is so that the pruner throws away all non-selecting, non-restricting
   # multijoins (since we def. do not care about those inside the subquery)
-  my $inner_from = $self->_prune_unused_joins ($from, $inner_select, $where, {
-    group_by => ['dummy'], %$inner_attrs,
-  });
 
-  # if a multi-type join was needed in the subquery - add a group_by to simulate the
-  # collapse in the subq
-  $inner_attrs->{group_by} ||= $inner_select
-    if first { ! $_->[0]{-is_single} } (@{$inner_from}[1 .. $#$inner_from]);
+  my $inner_subq = do {
 
-  # generate the subquery
-  my $subq = $self->_select_args_to_query (
-    $inner_from,
-    $inner_select,
-    $where,
-    $inner_attrs,
-  );
+    # must use it here regardless of user requests
+    local $self->{_use_join_optimizer} = 1;
 
-  my $subq_joinspec = {
-    -alias => $attrs->{alias},
-    -source_handle => $inner_from->[0]{-source_handle},
-    $attrs->{alias} => $subq,
+    my $inner_from = $self->_prune_unused_joins ($from, $inner_select, $where, {
+      group_by => ['dummy'], %$inner_attrs,
+    });
+
+    my $inner_aliastypes =
+      $self->_resolve_aliastypes_from_select_args( $inner_from, $inner_select, $where, $inner_attrs );
+
+    # we need to simulate collapse in the subq if a multiplying join is pulled
+    # by being a non-selecting restrictor
+    if (
+      ! $inner_attrs->{group_by}
+        and
+      first {
+        $inner_aliastypes->{restricting}{$_}
+          and
+        ! $inner_aliastypes->{selecting}{$_}
+      } ( keys %{$inner_aliastypes->{multiplying}||{}} )
+    ) {
+      my $unprocessed_order_chunks;
+      ($inner_attrs->{group_by}, $unprocessed_order_chunks) = $self->_group_over_selection (
+        $inner_from, $inner_select, $inner_attrs->{order_by}
+      );
+
+      $self->throw_exception (
+        'A required group_by clause could not be constructed automatically due to a complex '
+      . 'order_by criteria. Either order_by columns only (no functions) or construct a suitable '
+      . 'group_by by hand'
+      )  if $unprocessed_order_chunks;
+    }
+
+    # we already optimized $inner_from above
+    local $self->{_use_join_optimizer} = 0;
+
+    # generate the subquery
+    $self->_select_args_to_query (
+      $inner_from,
+      $inner_select,
+      $where,
+      $inner_attrs,
+    );
   };
 
   # Generate the outer from - this is relatively easy (really just replace
@@ -145,8 +179,13 @@ sub _adjust_select_args_for_complex_prefetch {
     $j = [ $j ] unless ref $j eq 'ARRAY'; # promote the head-from to an AoH
 
     if ($j->[0]{-alias} eq $attrs->{alias}) { # time to swap
+
       push @outer_from, [
-        $subq_joinspec,
+        {
+          -alias => $attrs->{alias},
+          -rsrc => $j->[0]{-rsrc},
+          $attrs->{alias} => $inner_subq,
+        },
         @{$j}[1 .. $#$j],
       ];
       last; # we'll take care of what's left in $from below
@@ -156,30 +195,52 @@ sub _adjust_select_args_for_complex_prefetch {
     }
   }
 
-  # scan the from spec against different attributes, and see which joins are needed
+  # scan the *remaining* from spec against different attributes, and see which joins are needed
   # in what role
   my $outer_aliastypes =
     $self->_resolve_aliastypes_from_select_args( $from, $outer_select, $where, $outer_attrs );
 
+  # unroll parents
+  my ($outer_select_chain, $outer_restrict_chain) = map { +{
+    map { $_ => 1 } map { values %$_} map { @$_ } values %{ $outer_aliastypes->{$_} || {} }
+  } } qw/selecting restricting/;
+
   # see what's left - throw away if not selecting/restricting
-  # also throw in a group_by if restricting to guard against
-  # cross-join explosions
-  #
+  # also throw in a group_by if a non-selecting multiplier,
+  # to guard against cross-join explosions
+  my $need_outer_group_by;
   while (my $j = shift @$from) {
     my $alias = $j->[0]{-alias};
 
-    if ($outer_aliastypes->{selecting}{$alias}) {
-      push @outer_from, $j;
+    if (
+      $outer_select_chain->{$alias}
+    ) {
+      push @outer_from, $j
     }
-    elsif ($outer_aliastypes->{restricting}{$alias}) {
+    elsif ($outer_restrict_chain->{$alias}) {
       push @outer_from, $j;
-      $outer_attrs->{group_by} ||= $outer_select unless $j->[0]{-is_single};
+      $need_outer_group_by ||= $outer_aliastypes->{multiplying}{$alias} ? 1 : 0;
     }
   }
 
   # demote the outer_from head
   $outer_from[0] = $outer_from[0][0];
 
+  if ($need_outer_group_by and ! $outer_attrs->{group_by}) {
+
+    my $unprocessed_order_chunks;
+    ($outer_attrs->{group_by}, $unprocessed_order_chunks) = $self->_group_over_selection (
+      \@outer_from, $outer_select, $outer_attrs->{order_by}
+    );
+
+    $self->throw_exception (
+      'A required group_by clause could not be constructed automatically due to a complex '
+    . 'order_by criteria. Either order_by columns only (no functions) or construct a suitable '
+    . 'group_by by hand'
+    ) if $unprocessed_order_chunks;
+
+  }
+
   # This is totally horrific - the $where ends up in both the inner and outer query
   # Unfortunately not much can be done until SQLA2 introspection arrives, and even
   # then if where conditions apply to the *right* side of the prefetch, you may have
@@ -219,8 +280,13 @@ sub _resolve_aliastypes_from_select_args {
       or next;
 
     $alias_list->{$al} = $j;
-    $aliases_by_type->{multiplying}{$al} = 1
-      if ref($_) eq 'ARRAY' and ! $j->{-is_single}; # not array == {from} head == can't be multiplying
+    $aliases_by_type->{multiplying}{$al} ||= $j->{-join_path}||[] if (
+      # not array == {from} head == can't be multiplying
+      ( ref($_) eq 'ARRAY' and ! $j->{-is_single} )
+        or
+      # a parent of ours is already a multiplier
+      ( grep { $aliases_by_type->{multiplying}{$_} } @{ $j->{-join_path}||[] } )
+    );
   }
 
   # get a column to source/alias map (including unqualified ones)
@@ -229,7 +295,11 @@ sub _resolve_aliastypes_from_select_args {
   # set up a botched SQLA
   my $sql_maker = $self->sql_maker;
 
-  local $sql_maker->{having_bind};  # these are throw away results
+  # these are throw away results, do not pollute the bind stack
+  local $sql_maker->{select_bind};
+  local $sql_maker->{where_bind};
+  local $sql_maker->{group_bind};
+  local $sql_maker->{having_bind};
 
   # we can't scan properly without any quoting (\b doesn't cut it
   # everywhere), so unless there is proper quoting set - use our
@@ -241,7 +311,9 @@ sub _resolve_aliastypes_from_select_args {
   local $sql_maker->{name_sep} = $sql_maker->{name_sep};
 
   unless (defined $sql_maker->{quote_char} and length $sql_maker->{quote_char}) {
-    $sql_maker->{quote_char} = "\x00";
+    $sql_maker->{quote_char} = ["\x00", "\xFF"];
+    # if we don't unset it we screw up retarded but unfortunately working
+    # 'MAX(foo.bar)' => { '>', 3 }
     $sql_maker->{name_sep} = '';
   }
 
@@ -256,8 +328,8 @@ sub _resolve_aliastypes_from_select_args {
       }),
     ],
     selecting => [
-      $self->_extract_order_columns ($attrs->{order_by}, $sql_maker),
       $sql_maker->_recurse_fields ($select),
+      ( map { $_->[0] } $self->_extract_order_criteria ($attrs->{order_by}, $sql_maker) ),
     ],
   };
 
@@ -275,7 +347,8 @@ sub _resolve_aliastypes_from_select_args {
 
     for my $type (keys %$to_scan) {
       for my $piece (@{$to_scan->{$type}}) {
-        $aliases_by_type->{$type}{$alias} = 1 if ($piece =~ $al_re);
+        $aliases_by_type->{$type}{$alias} ||= $alias_list->{$alias}{-join_path}||[]
+          if ($piece =~ $al_re);
       }
     }
   }
@@ -289,7 +362,10 @@ sub _resolve_aliastypes_from_select_args {
 
     for my $type (keys %$to_scan) {
       for my $piece (@{$to_scan->{$type}}) {
-        $aliases_by_type->{$type}{$colinfo->{$col}{-source_alias}} = 1 if ($piece =~ $col_re);
+        if ($piece =~ $col_re) {
+          my $alias = $colinfo->{$col}{-source_alias};
+          $aliases_by_type->{$type}{$alias} ||= $alias_list->{$alias}{-join_path}||[];
+        }
       }
     }
   }
@@ -297,23 +373,62 @@ sub _resolve_aliastypes_from_select_args {
   # Add any non-left joins to the restriction list (such joins are indeed restrictions)
   for my $j (values %$alias_list) {
     my $alias = $j->{-alias} or next;
-    $aliases_by_type->{restricting}{$alias} = 1 if (
+    $aliases_by_type->{restricting}{$alias} ||= $j->{-join_path}||[] if (
       (not $j->{-join_type})
         or
       ($j->{-join_type} !~ /^left (?: \s+ outer)? $/xi)
     );
   }
 
-  # mark all join parents as mentioned
-  # (e.g.  join => { cds => 'tracks' } - tracks will need to bring cds too )
-  for my $type (keys %$aliases_by_type) {
-    for my $alias (keys %{$aliases_by_type->{$type}}) {
-      $aliases_by_type->{$type}{$_} = 1
-        for (map { values %$_ } @{ $alias_list->{$alias}{-join_path} || [] });
+  return $aliases_by_type;
+}
+
+# This is the engine behind { distinct => 1 }
+sub _group_over_selection {
+  my ($self, $from, $select, $order_by) = @_;
+
+  my $rs_column_list = $self->_resolve_column_info ($from);
+
+  my (@group_by, %group_index);
+
+  # the logic is: if it is a { func => val } we assume an aggregate,
+  # otherwise if \'...' or \[...] we assume the user knows what is
+  # going on thus group over it
+  for (@$select) {
+    if (! ref($_) or ref ($_) ne 'HASH' ) {
+      push @group_by, $_;
+      $group_index{$_}++;
+      if ($rs_column_list->{$_} and $_ !~ /\./ ) {
+        # add a fully qualified version as well
+        $group_index{"$rs_column_list->{$_}{-source_alias}.$_"}++;
+      }
     }
   }
 
-  return $aliases_by_type;
+  # add any order_by parts that are not already present in the group_by
+  # we need to be careful not to add any named functions/aggregates
+  # i.e. order_by => [ ... { count => 'foo' } ... ]
+  my @leftovers;
+  for ($self->_extract_order_criteria($order_by)) {
+    # only consider real columns (for functions the user got to do an explicit group_by)
+    if (@$_ != 1) {
+      push @leftovers, $_;
+      next;
+    }
+    my $chunk = $_->[0];
+    my $colinfo = $rs_column_list->{$chunk} or do {
+      push @leftovers, $_;
+      next;
+    };
+
+    $chunk = "$colinfo->{-source_alias}.$chunk" if $chunk !~ /\./;
+    push @group_by, $chunk unless $group_index{$chunk}++;
+  }
+
+  return wantarray
+    ? (\@group_by, (@leftovers ? \@leftovers : undef) )
+    : \@group_by
+  ;
 }
 
 sub _resolve_ident_sources {
@@ -341,8 +456,8 @@ sub _resolve_ident_sources {
         $tabinfo = $_->[0];
       }
 
-      $alias2source->{$tabinfo->{-alias}} = $tabinfo->{-source_handle}->resolve
-        if ($tabinfo->{-source_handle});
+      $alias2source->{$tabinfo->{-alias}} = $tabinfo->{-rsrc}
+        if ($tabinfo->{-rsrc});
     }
   }
 
@@ -360,7 +475,7 @@ sub _resolve_column_info {
   my ($self, $ident, $colnames) = @_;
   my ($alias2src, $root_alias) = $self->_resolve_ident_sources($ident);
 
-  my (%return, %seen_cols, @auto_colnames);
+  my (%seen_cols, @auto_colnames);
 
   # compile a global list of column names, to be able to properly
   # disambiguate unqualified column names (if at all possible)
@@ -377,25 +492,29 @@ sub _resolve_column_info {
     grep { @{$seen_cols{$_}} == 1 } (keys %seen_cols),
   ];
 
-  COLUMN:
+  my (%return, $colinfos);
   foreach my $col (@$colnames) {
-    my ($alias, $colname) = $col =~ m/^ (?: ([^\.]+) \. )? (.+) $/x;
+    my ($source_alias, $colname) = $col =~ m/^ (?: ([^\.]+) \. )? (.+) $/x;
 
-    unless ($alias) {
-      # see if the column was seen exactly once (so we know which rsrc it came from)
-      if ($seen_cols{$colname} and @{$seen_cols{$colname}} == 1) {
-        $alias = $seen_cols{$colname}[0];
-      }
-      else {
-        next COLUMN;
-      }
-    }
+    # if the column was seen exactly once - we know which rsrc it came from
+    $source_alias ||= $seen_cols{$colname}[0]
+      if ($seen_cols{$colname} and @{$seen_cols{$colname}} == 1);
 
-    my $rsrc = $alias2src->{$alias};
-    $return{$col} = $rsrc && {
-      %{$rsrc->column_info($colname)},
+    next unless $source_alias;
+
+    my $rsrc = $alias2src->{$source_alias}
+      or next;
+
+    $return{$col} = {
+      %{
+          ( $colinfos->{$source_alias} ||= $rsrc->columns_info )->{$colname}
+            ||
+          $self->throw_exception(
+            "No such column '$colname' on source " . $rsrc->source_name
+          );
+      },
       -result_source => $rsrc,
-      -source_alias => $alias,
+      -source_alias => $source_alias,
     };
   }
 
@@ -472,72 +591,7 @@ sub _inner_join_to_node {
   return \@new_from;
 }
 
-# Most databases do not allow aliasing of tables in UPDATE/DELETE. Thus
-# a condition containing 'me' or other table prefixes will not work
-# at all. What this code tries to do (badly) is introspect the condition
-# and remove all column qualifiers. If it bails out early (returns undef)
-# the calling code should try another approach (e.g. a subquery)
-sub _strip_cond_qualifiers {
-  my ($self, $where) = @_;
-
-  my $cond = {};
-
-  # No-op. No condition, we're updating/deleting everything
-  return $cond unless $where;
-
-  if (ref $where eq 'ARRAY') {
-    $cond = [
-      map {
-        my %hash;
-        foreach my $key (keys %{$_}) {
-          $key =~ /([^.]+)$/;
-          $hash{$1} = $_->{$key};
-        }
-        \%hash;
-      } @$where
-    ];
-  }
-  elsif (ref $where eq 'HASH') {
-    if ( (keys %$where) == 1 && ( (keys %{$where})[0] eq '-and' )) {
-      $cond->{-and} = [];
-      my @cond = @{$where->{-and}};
-       for (my $i = 0; $i < @cond; $i++) {
-        my $entry = $cond[$i];
-        my $hash;
-        my $ref = ref $entry;
-        if ($ref eq 'HASH' or $ref eq 'ARRAY') {
-          $hash = $self->_strip_cond_qualifiers($entry);
-        }
-        elsif (! $ref) {
-          $entry =~ /([^.]+)$/;
-          $hash->{$1} = $cond[++$i];
-        }
-        else {
-          $self->throw_exception ("_strip_cond_qualifiers() is unable to handle a condition reftype $ref");
-        }
-        push @{$cond->{-and}}, $hash;
-      }
-    }
-    else {
-      foreach my $key (keys %$where) {
-        if ($key eq '-or' && ref $where->{$key} eq 'ARRAY') {
-          $cond->{$key} = $self->_strip_cond_qualifiers($where->{$key});
-        }
-        else {
-          $key =~ /([^.]+)$/;
-          $cond->{$1} = $where->{$key};
-        }
-      }
-    }
-  }
-  else {
-    return undef;
-  }
-
-  return $cond;
-}
-
-sub _extract_order_columns {
+sub _extract_order_criteria {
   my ($self, $order_by, $sql_maker) = @_;
 
   my $parser = sub {
@@ -547,8 +601,9 @@ sub _extract_order_columns {
       unless wantarray;
 
     my @chunks;
-    for my $chunk (map { ref $_ ? @$_ : $_ } ($sql_maker->_order_by_chunks ($order_by) ) ) {
-      $chunk =~ s/\s+ (?: ASC|DESC ) \s* $//ix;
+    for ($sql_maker->_order_by_chunks ($order_by) ) {
+      my $chunk = ref $_ ? $_ : [ $_ ];
+      $chunk->[0] =~ s/\s+ (?: ASC|DESC ) \s* $//ix;
       push @chunks, $chunk;
     }
 
